(module asplevels-ex-test-core scheme
  (require redex)
  (require "../asplevels-ex-core.rkt")
  (provide asplevels-ex-test-core)
  
  (define (asplevels-ex-test-core λreduction-rules)
    
    (begin
      (display "Running core tests")
      (newline)
      
      ;; Testing let
      (test-->>
       λreduction-rules
       (term 
        (0 (stack) (asps)
           (let ((x 3) (y (+ 2 2))) (+ x y))))
       (term (0 (stack) (asps) 7)))
      
      ;; Testing lists and if
      
      (test-->>
       λreduction-rules
       (term
        (0 (stack) (asps)
           (car (cdr (cons 3 (list 1 2))))))
       (term (0 (stack) (asps) 1)))
      
      (test-->>
       ; (traces
       λreduction-rules
       (term
        (0 (stack) (asps)
           (empty? (list))))
       ;)
       (term (0 (stack) (asps) #t)))
      
      (test-->>
       λreduction-rules
       (term
        (0 (stack) (asps)
           (empty? (list 1 3))))
       (term (0 (stack) (asps) #f)))
      
      (test-->>
       λreduction-rules
       (term
        (0 (stack) (asps)
           (empty? (cdr (cdr (cons 2 (cons 1 (list))))))))
       (term (0 (stack) (asps) #t)))
      
      (test-->>
       λreduction-rules
       (term
        (0 (stack) (asps)
           (if (empty? (cdr (cdr (cons 2 (cons 1 (list))))))
               1
               2)))
       (term (0 (stack) (asps) 1)))
      
      (test-->>
       λreduction-rules
       (term
        (0 (stack) (asps)
           (if (empty? (cdr (cons 2 (list 1))))
               1
               2)))
       (term (0 (stack) (asps) 2)))
      
      (test-->>
       λreduction-rules
       (term
        (0 (stack) (asps)
           (if (cdr (cdr (list 1 2)))
               1
               2)))
       (term (0 (stack) (asps) 1)))
      
      ;; Testing eq?
      (test-->>
       λreduction-rules
       (term 
        (0 (stack) (asps)
           (if (eq? (+ 2 3) 0)
               1
               0)))
       (term (0 (stack) (asps) 0)))
      
      
      (test-->>
       λreduction-rules
       (term 
        (0 (stack) (asps)
           (if (eq? (+ 2 3) (+ 1 4))
               1
               0)))
       (term (0 (stack) (asps) 1)))
      
      (test-->>
       λreduction-rules
       (term 
        (0 (stack) (asps)
           (if (eq? "hola" "hola")
               1
               0)))
       (term (0 (stack) (asps) 1)))
      
      ; function identity check
      (test-->>
       λreduction-rules 
       (term
        (0 (stack) (asps)
           (let ((f (λ (x) x)))
             (eq? f f))))
       (term (0 (stack) (asps) #t)))
      
      (test-->>
       ;(traces 
       λreduction-rules
       (term
        (0 (stack) (asps)
           ((λ (f)
              (if (eq? (f 4) 5)
                  ((λ () #t))
                  ((λ () #f))))
            ((λ (x) 
               (λ (y) (+ x y))) 2))
           ))
       ;)
       (term (0 (stack) (asps) #f)))
      
      ;; Testing aspects
      (test-->>
       ; (traces 
       λreduction-rules
       (term
        (0 (stack) (asps (0 (λ r1 -1 (jpt) (list))
                            (λ r2 -1 (q c x y) (q x y))))
           ((λ (x y) (+ x y))
            2 3)))
       ;)
       (term (0 (stack) (asps (0 (λ r1 -1 (jpt) (list))
                                 (λ r2 -1 (q c x y) (q x y)))) 5)))
      
      (test-->>
       ; (traces 
       λreduction-rules
       (term
        (0 (stack) (asps (0 (λ r1 -1 (jpt) (list))
                            (λ r2 -1 (q c x y) (- x y))))
           ((λ (x y) (+ x y))
            2 3)))
       ;)
       (term (0 (stack) (asps (0 (λ r1 -1 (jpt) (list))
                                 (λ r2 -1 (q c x y) (- x y)))) 
                -1)))
      
      ;; 2 aspects, each changing an arg
      (test-->>
       ; (traces 
       λreduction-rules
       (term
        (0 (stack) (asps) 
           (begin (deploy (λ r1 -1 (jpt) (list 10))
                          (λ r2 -1 (q c x y) (q (car c) y)))
                  (deploy (λ r1 -1 (jpt) (list 1))
                          (λ r2 -1 (q c x y) (q x (car c))))
                  
                  ((λ (x y) (+ x y))
                   2 3))))
       ;)
       (term (0 (stack) (asps (0 (λ r1 -1 (jpt) (list 1))
                                 (λ r2 -1 (q c x y) (q x (car c))))
                              (0 (λ r1 -1 (jpt) (list 10))
                                 (λ r2 -1 (q c x y) (q (car c) y)))) 
                11)))
      
      ;; 1 aspect on each level (2)
      (test-predicate
       (λ (res)
         (redex-match λal (0 (stack) A_1 7)
                      (car res)))  
       (apply-reduction-relation* 
        λreduction-rules
        (term
         (0 (stack) (asps) 
            (let ((pc0 (λ (jpt) (list 1))))
              (begin 
                ;; level 1 -> replaces binding returned by pointcut of level 0 aspect
                (up (deploy (λ (jpt) (if (eq? (car (cdr jpt)) pc0)
                                         (list)
                                         #f))
                            (λ (q c x) (list 5))))
                
                ;; level 0 -> replaces 2nd parameter of base call
                (deploy pc0
                        (λ (q c x y) (q x (car c))))
                
                ((λ (x y) (+ x y))
                 2 3)))))))
      
      ;; 1 aspect on each level (3 / same as above + level2 aspect overriding advice)
      ;(test-->>
      (test-predicate
       (λ (res)
         (redex-match λal (0 (stack) A_1 12)
                      (car res)))
       (apply-reduction-relation*
        λreduction-rules
        (term
         (0 (stack) (asps)
            (let ((adv1 (λ (q c x) (list 5)))
                  (pc0  (λ (jpt) (list 1))))
              (begin 
                ;; level 2 -> replaces value returned by advice of level 1 aspect
                (up (up (deploy (λ (jpt) (if (eq? (car (cdr jpt)) adv1)
                                             (list)
                                             #f))
                                (λ (p c bp bc bx) (list 10)))))
                
                ;; level 1 -> replaces binding returned by pointcut of level 0 aspect
                (up (deploy (λ (jpt) (if (eq? (car (cdr jpt)) pc0) 
                                         (list)
                                         #f))
                            adv1))
                
                ;; level 0 -> replaces 2nd parameter of base call
                (deploy pc0 (λ (q c x y) (q x (car c))))
                
                ((λ (x y) (+ x y))
                 2 3)))))))
      
      ;; level-capturing functions
      (test-predicate
       (λ (res)
         (redex-match λal (0 (stack) A_1 4)
                      (car res)))
       (apply-reduction-relation*
        λreduction-rules
        (term (0 (stack) (asps)
                 (let ((f (λ (x) (+ x x)))
                       (task (up (λs (x) (f x))))) ;; task runs at level 1
                   (begin 
                     (deploy (λ (jpt) (eq? (car (cdr jpt)) f)) ;; should not see f apply
                             (λ (q c x) 5))
                     (task 2)))))))
      
      (test-predicate
       (λ (res)
         (redex-match λal (0 (stack) A_1 5)
                      (car res)))
       (apply-reduction-relation*
        λreduction-rules
        (term (0 (stack) (asps)
                 (let ((f (λ (x) (+ x x)))
                       (task (up (λs (x) (f x))))) ;; runs at level 1
                   (begin 
                     (up (deploy (λ (jpt) (eq? (car (cdr jpt)) f)) ;; up: sees f apply
                                 (λ (q c x) 5)))
                     (task 2)))))))
      
      (test-results))))

